#include "typedef.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "object.h"
#include "opcode.h"
#include "class.h"
#include "detect.h"
#include "util.h"

extern int	g_report_java;
extern int	g_report_pcode;
extern int	g_verbose;
extern char	*g_html_dirP;
extern int	g_comments;
extern int	g_min_match_sequence;
extern int	g_mismatch_cost;
extern int	g_same_cost;
extern int	g_match_lines;
extern int  g_comparisons;

static	Cclass			*class1P;		// Class 1st side looking at
static	Cclass			*class2P;		// Class 2nd side looking at
static	method_infoT	*method1P;		// Method 1st side looking at
static	method_infoT	*method2P;		// Method 2nd side looking at
static	method_infoT	*end_method1P;	// End of 1st side methods
static	method_infoT	*end_method2P;	// End of 2nd side methods
static	pcodeT			*start_pcode1P;	// Comparing from 1st side
static	pcodeT			*start_pcode2P;	// Comparing from 2nd side
static	pcodeT			*end_pcode1P;	// May not compare to here on 1st side
static	pcodeT			*end_pcode2P;	// May not compare to here on 2nd side

static int		started       = 0;
static int		hypertext_cnt = 0;
extern FILE		*stdoutF;
extern FILE		*redirectF;

static Cclass	*last_classP = 0;
static FILE		*classF      = 0;
static FILE		*cloneF      = 0;

void
start_report(void)
{
	if (g_html_dirP) {
		fputs("<p>\n<table>\n", redirectF);
}	}

static void
report_clone(pcodeT	*end_code1P, pcodeT *end_code2P)
{
	static	u2T				*class_name1P;
	static	FILE			*outF;
	static	int				lineno1, lineno2, next_lineno1, next_lineno2;
	static	int				different1, different2;
	static	int				report_java;
	static	const opcodeT	*opcode1P, *opcode2P;

	char filename[2048];

	pcodeT	*at_code1P, *at_code2P;

	if (g_html_dirP) {
		if (last_classP != class1P) {
			if (classF) {
				fprintf(classF,"</table>\n<p>\nEnd of report</body>\n</html>\n");
				fclose(classF);
			}
			sprintf(filename,"%s/%d.html", g_html_dirP, ++hypertext_cnt);
			classF = fopen(filename, "w");
			if (!classF) {
				fprintf(stderr, "Can't open %s\n", filename);
				exit(1);
			}
			last_classP = class1P;

			stdoutF      = classF;
			class_name1P = class1P->get_classname(class1P->m_this_class);

			fprintf(stdoutF, "<html>\n<title>Clones in ");
			dump_classname(class_name1P);
			fprintf(stdoutF, "</title>\n<body>\n<h3>Clones in ");
			dump_classname(class_name1P);
			fprintf(stdoutF, "</h3>\n<table>\n");
			
			stdoutF = redirectF;
			fprintf(stdoutF, "<tr><td><a href=\"%d.html\">", hypertext_cnt);
			dump_classname(class1P->get_classname(class1P->m_this_class));
			fprintf(stdoutF, "</a></td></tr>\n");
		}

		cloneF = 0;
		if (g_report_java || g_report_pcode || g_comments) {
			sprintf(filename,"%s/%d.html", g_html_dirP, ++hypertext_cnt);
			cloneF = fopen(filename, "w");
			if (!cloneF) {
				fprintf(stderr, "Can't open %s\n", filename);
				exit(1);
		}	}
		
		stdoutF = classF;
		fputs("<tr><td><i>", stdoutF);
		if (cloneF) {
			fprintf(stdoutF, "<a href=\"%d.html\">", hypertext_cnt);
		}
		dump_access(method1P->m_access_flags);
		dumpUnicode(method1P->m_nameP);
		class1P->dump_descriptor(method1P->m_descriptorP, method1P, (method1P->m_access_flags & 8));
		if (cloneF) {
			fputs("</a>", stdoutF);
		}
		fprintf(stdoutF, "</i> <font color=\"red\">[%d]</font></td>\n    <td><i>", end_code1P - start_pcode1P);
		dump_access(method2P->m_access_flags);
		dump_classname(class2P->get_classname(class2P->m_this_class));
		fputc('.', stdoutF);
		dumpUnicode(method2P->m_nameP);
		class2P->dump_descriptor(method2P->m_descriptorP, method2P, (method2P->m_access_flags & 8));
		fprintf(stdoutF, "</i> <font color=\"blue\">[%d]</font></td>\n", end_code2P - start_pcode2P);

		fputs("</i></td>\n</tr>\n", stdoutF);
		if (!cloneF) {
			stdoutF = redirectF;
			return;
		}
		stdoutF = cloneF;
	} 
	fputs("<p>\n<table>\n", stdoutF);
	fputs("<tr><td><b>", stdoutF);
	dump_classname(class1P->get_classname(class1P->m_this_class));
	fputs("</b></td>\n    <td><b>", stdoutF);
	dump_classname(class2P->get_classname(class2P->m_this_class));
	fputs("</b></td>\n</tr>\n", stdoutF);
	fprintf(stdoutF, "<tr><td><i>");
	dump_access(method1P->m_access_flags);
	dumpUnicode(method1P->m_nameP);
	class1P->dump_descriptor(method1P->m_descriptorP, method1P, (method1P->m_access_flags & 8));
	fprintf(stdoutF, "</i> <font color=\"red\">[%d]</font></td>\n    <td><i>", end_code1P - start_pcode1P);
	dump_access(method2P->m_access_flags);
	dumpUnicode(method2P->m_nameP);
	class2P->dump_descriptor(method2P->m_descriptorP, method2P, (method2P->m_access_flags & 8));
	fprintf(stdoutF, "</i> <font color=\"blue\">[%d]</font></td>\n</tr>\n", end_code2P - start_pcode2P);

	if (g_report_java || g_report_pcode || g_comments) {

		fputs("<tr><td><hr></hr></td><td><hr></hr></td></tr>", stdoutF); 

		at_code1P = start_pcode1P;
		at_code2P = start_pcode2P;

		if (report_java = g_report_java) {

			if (!class1P->m_sourcePP) {
				class1P->load_source();
			}
			if (!class2P->m_sourcePP) {
				class2P->load_source();
			}
			if (!class1P->m_source_lines && !class2P->m_source_lines) {
				// Can open neither source
				report_java = 0;
			}
			next_lineno1 = ((at_code1P < end_code1P) ? at_code1P->m_lineno : -1);
			next_lineno2 = ((at_code2P < end_code2P) ? at_code2P->m_lineno : -1);
			different1   = 0;
			different2   = 0;
		}

		for (; at_code1P < end_code1P || at_code2P < end_code2P; ) {

			if (g_comments) {
				opcode1P = opcodes + at_code1P->m_pcode;
				opcode2P = opcodes + at_code2P->m_pcode;

				fputs("<tr><td>", stdoutF);
				if (end_code1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					fputs("<tr><td></td>\n    <td><font color=\"blue\">", stdoutF);
					dumpString(opcode2P->commentP);
					fputs("</font></td>\n</tr>\n", stdoutF);
				} else if (!at_code1P->m_matchP) {
					fputs("<tr><td><font color=\"red\">", stdoutF);
					dumpString(opcode1P->commentP);
					fputs("</font></td>\n    <td>", stdoutF);
				
					if (at_code2P < end_code2P && !at_code2P->m_matchP) {
						fputs("<font color=\"blue\">", stdoutF);
						dumpString(opcode2P->commentP);
						fputs("</font>", stdoutF);
					}
					fputs("</td>\n</tr>\n", stdoutF);
				} else {
					fputs("<tr><td>", stdoutF);
					dumpString(opcode1P->commentP);
					fputs("</td>\n    <td>", stdoutF);
					dumpString(opcode2P->commentP);
					fputs("</td>\n</tr>\n", stdoutF);
			}	} 

			if (g_report_pcode) {
				fputs("<tr><td>", stdoutF);
				if (end_code1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					fputs("<tr><td></td>\n    <td><font color=\"blue\">", stdoutF);
					class2P->dump_pcode(method2P, at_code2P++);
					fputs("</font></td>\n</tr>\n", stdoutF);
					different2 = 1;
				} else if (!at_code1P->m_matchP) {
					fputs("<tr><td><font color=\"red\">", stdoutF);
					class1P->dump_pcode(method1P, at_code1P++);
					fputs("</font></td>\n    <td>", stdoutF);
					different1 = 1;
				
					if (at_code2P < end_code2P && !at_code2P->m_matchP) {
						fputs("<font color=\"blue\">", stdoutF);
						class2P->dump_pcode(method2P, at_code2P++);
						fputs("</font>", stdoutF);
						different2 = 1;
					}
					fputs("</td>\n</tr>\n", stdoutF);
				} else {
					fputs("<tr><td>", stdoutF);
					class1P->dump_pcode(method1P, at_code1P++);
					fputs("</td>\n    <td>", stdoutF);
					class2P->dump_pcode(method2P, at_code2P++);
					fputs("</td>\n</tr>\n", stdoutF);
				} 
			} else {
				if (end_code1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					at_code2P++;
					different2 = 1;
				} else if (!at_code1P->m_matchP) {
					at_code1P++;
					different1 = 1;
				
					if (at_code2P < end_code2P && !at_code2P->m_matchP) {
						at_code2P++;
					}
				} else {
					at_code1P++;
					at_code2P++;
			}	} 

			if (report_java) {

				lineno1      = next_lineno1;
				lineno2      = next_lineno2;
				next_lineno1 = ((at_code1P < end_code1P) ? at_code1P->m_lineno : -1);
				next_lineno2 = ((at_code2P < end_code2P) ? at_code2P->m_lineno : -1);

				if (lineno1 == next_lineno1) {
					if (!class2P->m_source_lines || lineno2 == next_lineno2) {
						continue;
					}
					fputs("<tr><td></td>\n    <td><font color=\"blue\">", stdoutF);
					class2P->showline(lineno2);
					fputs("</font></td>\n</tr>\n", stdoutF);
				} else if (lineno2 == next_lineno2) {
					if (!class1P->m_source_lines || lineno1 == next_lineno1) {
						continue;
					}
					fputs("<tr><td><font color=\"red\">", stdoutF);
					class1P->showline(lineno1);
					fputs("</font></td>\n    <td></td></tr>\n", stdoutF);
				} else {
					fputs("<tr><td>", stdoutF);
					if (class1P->m_source_lines) {
						if (!different1) {
							class1P->showline(lineno1);
						} else {
							fputs("<font color=\"red\">", stdoutF);
							class1P->showline(lineno1);
							fputs("</font>", stdoutF);
					}	}
					fputs("</td><td>", stdoutF);
					if (class2P->m_source_lines) {
						if (!different2) {
							class2P->showline(lineno2);
						} else {
							fputs("<font color=\"blue\">", stdoutF);
							class2P->showline(lineno2);
							fputs("</font>", stdoutF);
					}	}
					fputs("</td></tr>\n", stdoutF);
				}
				different1 = 0;
				different2 = 0;
			}
		}
	}
	fputs("</table>\n", stdoutF);
	if (g_html_dirP) {
		fputs("<p>End of report</body>\n</html>\n", stdoutF);
		fclose(stdoutF);
		stdoutF = redirectF;
	}
}

void
finalize_report(void)
{
	if (classF) {
		fprintf(classF,"</table>\n<p>\n<p>End of report</body>\n</html>\n");
		fclose(classF);
	}
}

static int
same_branch(int	from1, int from2, int shift1, int shift2)
{
	int	to1, to2;

	if (shift1 < 0) {
		pcodeT	*pcode1P;
		pcodeT	*matchP;

		if (shift2 >= 0) {
			return g_mismatch_cost;
		}
		to2 = from2 + shift2;
		if (to2 < (int) start_pcode2P->m_pc) {
			return g_mismatch_cost;
		}
		to1 = from1 + shift1;
		for (pcode1P = start_pcode1P; pcode1P->m_pc != to1 || !(matchP = pcode1P->m_matchP); ++pcode1P) {
			if (to1 < (int) pcode1P->m_pc) {
				return g_mismatch_cost;
		}	}

		if (to2 != (int) matchP->m_pc) {
			return g_mismatch_cost;
		}
	} else {
		if (shift2 < 0) {
			return g_mismatch_cost;
		}
	}
	return g_same_cost;
}

static int
same_pcode(pcodeT *code1P, pcodeT *code2P) 
{
	int	pcode1;

	pcode1 = code1P->m_pcode;
	if (pcode1 != code2P->m_pcode) {
		return g_mismatch_cost;
	}

	switch (pcode1) {
	case ILOAD:
	case ILOAD_0:
	case ILOAD_1:
	case ILOAD_2:
	case ILOAD_3:
	case LLOAD:
	case LLOAD_0:
	case LLOAD_1:
	case LLOAD_2:
	case LLOAD_3:
	case FLOAD:
	case FLOAD_0:
	case FLOAD_1:
	case FLOAD_2:
	case FLOAD_3:
	case DLOAD:
	case DLOAD_0:
	case DLOAD_1:
	case DLOAD_2:
	case DLOAD_3:
	case ALOAD:
	case ALOAD_0:
	case ALOAD_1:
	case ALOAD_2:
	case ALOAD_3:

	case ISTORE:
	case ISTORE_0:
	case ISTORE_1:
	case ISTORE_2:
	case ISTORE_3:
	case LSTORE:
	case LSTORE_0:
	case LSTORE_1:
	case LSTORE_2:
	case LSTORE_3:
	case FSTORE:
	case FSTORE_0:
	case FSTORE_1:
	case FSTORE_2:
	case FSTORE_3:
	case DSTORE:
	case DSTORE_0:
	case DSTORE_1:
	case DSTORE_2:
	case DSTORE_3:
	case ASTORE:
	case ASTORE_0:
	case ASTORE_1:
	case ASTORE_2:
	case ASTORE_3:
	{
		varT	*var1P, *var2P;
		
		var1P = (varT *) code1P->m_tableP;
		var2P = (varT *) code2P->m_tableP;

		if (var1P->m_nameP || var2P->m_nameP) {
			if (var1P->m_nameP == var2P->m_nameP) {
				break;
			}
			return g_mismatch_cost;
		}
		if (var1P->m_number == var2P->m_number) {
			break;
		}
		return g_mismatch_cost;
	}
	case IINC:
	{
		iincT	*var1P, *var2P;
		
		var1P = (iincT *) code1P->m_tableP;
		var2P = (iincT *) code2P->m_tableP;

		if (var1P->m_s4 != var2P->m_s4) {
			return g_mismatch_cost;
		}

		if (var1P->m_nameP && var2P->m_nameP) {
			if (var1P->m_nameP == var2P->m_nameP) {
				break;
			}
			return g_mismatch_cost;
		}
		if (var1P->m_number == var2P->m_number) {
			break;
		}
		return g_mismatch_cost;
	}
	case GETSTATIC:	
	case GETFIELD:
	case PUTSTATIC:
	case PUTFIELD:
	{
		fieldrefT	*fieldref1P, *fieldref2P;

		fieldref1P = (fieldrefT *) code1P->m_tableP;
		fieldref2P = (fieldrefT *) code2P->m_tableP;

		if (fieldref1P->m_classnameP != fieldref2P->m_classnameP ||
			fieldref1P->m_nameP      != fieldref2P->m_nameP) {
			return g_mismatch_cost;
		}
		break;
	}
	case INVOKEVIRTUAL:
	case INVOKESPECIAL:
	case INVOKESTATIC:
	{
		methodrefT	*methodref1P, *methodref2P;

		methodref1P = (methodrefT *) code1P->m_tableP;
		methodref2P = (methodrefT *) code2P->m_tableP;

		if (methodref1P->m_classnameP   != methodref2P->m_classnameP ||
			methodref1P->m_nameP        != methodref2P->m_nameP      ||
			methodref1P->m_descriptorP  != methodref2P->m_descriptorP) {
			return g_mismatch_cost;
		}
		break;
	}
	case INVOKEINTERFACE:
	{
		interfacerefT	*methodref1P, *methodref2P;

		methodref1P = (interfacerefT *) code1P->m_tableP;
		methodref2P = (interfacerefT *) code2P->m_tableP;

		if (methodref1P->m_classnameP   != methodref2P->m_classnameP ||
			methodref1P->m_nameP        != methodref2P->m_nameP      ||
			methodref1P->m_descriptorP  != methodref2P->m_descriptorP ||
			methodref1P->m_args         != methodref2P->m_args) {
			return g_mismatch_cost;
		}
		break;
	}
	case MULTIANEWARRAY:
	{
		arrayrefT	*arrayref1P, *arrayref2P;

		arrayref1P = (arrayrefT *) code1P->m_tableP;
		arrayref2P = (arrayrefT *) code2P->m_tableP;

		if (arrayref1P->m_classnameP != arrayref2P->m_classnameP ||
			arrayref1P->m_dimensions != arrayref2P->m_dimensions) {
			return g_mismatch_cost;
		}
		break;
	}
	case ANEWARRAY:				// anewarray
	case NEW:					// new
	case CHECKCAST:				// checkcast
	case INSTANCEOF:			// instanceof
	{
		classrefT	*classref1P, *classref2P;
		
		classref1P = (classrefT *) code1P->m_tableP;
		classref2P = (classrefT *) code2P->m_tableP;

		if (classref1P->m_classnameP != classref2P->m_classnameP) {
			return g_mismatch_cost;
		}
		break;
	}
	case LDC:
	case LDC_W:
	case LDC2_W:
	{
		cp_infoT	*cp1P, *cp2P;

		cp1P = (cp_infoT *) code1P->m_tableP;
		cp2P = (cp_infoT *) code2P->m_tableP;

		if (cp1P->tag != cp2P->tag) {
			return g_mismatch_cost;
		}
		switch (cp1P->tag) {
		case CONSTANT_Integer:
			if (cp1P->u.ival != cp2P->u.ival) {
				return g_mismatch_cost;
			}
			break;
		case CONSTANT_Float:
			if (cp1P->u.fval != cp2P->u.fval) {
				return g_mismatch_cost;
			}
			break;
		case CONSTANT_Long:
			if (cp1P->u.lval != cp2P->u.lval) {
				return g_mismatch_cost;
			}
			break;
		case CONSTANT_Double:
			if (cp1P->u.dval != cp2P->u.dval) {
				return g_mismatch_cost;
			}
			break;
		case CONSTANT_String:
		case CONSTANT_Class:
			// N.B. Even though the spec says that there
			//      is an indirection to the string via
			//		cpP->u.ref.index1 we've already removed
			//		it by the time we get here.
			if (cp1P->u.stringP != cp2P->u.stringP) {
				return g_mismatch_cost;
			}
			break;
		default:
			assert(0);
		}
		break;
	}
	case TABLESWITCH:
	{
		int	pc1, pc2;
		s4T	low1, low2, high1, high2;
		int	weight, seen, diff;
		u4T	*table1P, *table2P, *end_table1P, *end_table2P;

		table1P = code1P->m_tableP;
		table2P = code2P->m_tableP;
		pc1		= code1P->m_pc;
		pc2		= code2P->m_pc;

		// Compare default branches
		weight  = same_branch(pc1, pc2, *table1P, *table2P);
		seen    = 1;

		low1	= table1P[1];
		low2	= table2P[1];
		high1	= table1P[2];
		high2	= table2P[2];
		
		diff = (high1 - high2);
		if (diff < 0) {
			weight -= diff * g_mismatch_cost;
			seen   -= diff;
		} else {
			weight += diff * g_mismatch_cost;
			seen   += diff;
			high1   = high2;
		}
		table1P += 3;
		table2P += 3;
		end_table1P = table1P + (high1 - low1 + 1);
		end_table2P = table2P + (high1 - low2 + 1);

		diff = low1 - low2;
		if (diff < 0) {
			weight  -= diff * g_mismatch_cost;
			table1P -= diff;
			seen    -= diff;
		} else {
			weight  += diff * g_mismatch_cost;
			table2P += diff;
			seen    += diff;
		}
		seen += (end_table2P - table2P);

		for (; table2P < end_table2P; ++table1P, ++table2P) {
			weight += same_branch(pc1, pc2, *table1P, *table2P);
		}

		return (int) (((double) weight) / ((double) seen));
	}
	case LOOKUPSWITCH:
	{
		int	pc1, pc2;
		int	weight, seen;
		u4T	*table1P, *table2P, *end_table1P, *end_table2P;

		table1P = code1P->m_tableP;
		table2P = code2P->m_tableP;
		pc1		= code1P->m_pc;
		pc2		= code2P->m_pc;
		// Defaults
		weight = same_branch(pc1, pc2, *table1P, *table2P);
		seen   = 1;

		table1P    += 2;
		table2P    += 2;
		end_table1P = table1P + (table1P[-1] << 1);
		end_table2P = table2P + (table2P[-1] << 1);

		// The table match-offset pairs of the lookupswitch instruction must be sorted in increasing numerical order by match.

		while (table1P < end_table1P || table2P < end_table2P) {
			if (table1P < end_table1P) {
				if (end_table2P <= table2P || *table1P < *table2P) {
					weight  += g_mismatch_cost;
					++seen;
					table1P += 2;
					continue;
				}
				if (*table1P == *table2P) {
					weight += same_branch(pc1, pc2, table1P[1], table2P[1]);
					++seen;
					table1P += 2;
					table2P += 2;
					continue;
			}	}
			weight += g_mismatch_cost;
			++seen;
			table2P += 2;
		}

		return (int) (((double) weight) / ((double) seen));
	}
	case IFEQ:
	case IFNE:
	case IFLT:
	case IFGE:
	case IFGT:
	case IFLE:
	case IF_ICMPEQ:
	case IF_ICMPNE:
	case IF_ICMPLT:
	case IF_ICMPGE:
	case IF_ICMPGT:
	case IF_ICMPLE:
	case IF_ACMPEQ:
	case IF_ACMPNE:
	case IFNULL:
	case IFNOTNULL:
	case GOTO:
	case GOTO_W:
	case JSR:
	case JSR_W:
		return same_branch((int) code1P->m_pc, (int) code2P->m_pc, code1P->m_s4, code2P->m_s4);
	case LABEL:
		return same_branch((int) code1P->m_pc, (int) code2P->m_pc, (int) (code1P->m_u4 - code1P->m_pc), (int) (code2P->m_u4 - code2P->m_pc));
	case TRY_START:
	case TRY_END:
	case TRY_LABEL:
	{
		exceptionT	*exception1P = (exceptionT *) code1P->m_tableP;
		exceptionT	*exception2P = (exceptionT *) code2P->m_tableP;

		if (exception1P->m_catch_typeP != exception2P->m_catch_typeP) {
			return g_mismatch_cost;
		}
		break;
	}
	default:
		if (code1P->m_u4 != code2P->m_u4) {
			return g_mismatch_cost;
	}	}

	return g_same_cost;
}

static pcodeT *
detect_clone(void)
{
	static	matchesT	*matches_cacheP    = 0;
	static	matchesT	*matches_endP      = 0;

	pcodeT	*code1P, *code2P, *end_code1P, *end_code2P, *at_code1P, *at_code2P;
	pcodeT	*stack1P, *stack2P;
	int		best, weight, distance, ret;
	int		stack1, stack2;

	weight = same_pcode(start_pcode1P, start_pcode2P);
	if (weight > 0) {
		// Initially have to have a match
		end_code1P = code1P = start_pcode1P;
		end_code2P = code2P = start_pcode2P;
		code1P->m_matchP    = code2P;
		code2P->m_matchP    = code1P;
		if (g_match_lines == 1) {
			if (code1P->m_pop > 0) {
				return(0);
			}
			if (code2P->m_pop > 0) {
				return(0);
			}
			stack1 = code1P->m_push;
			stack2 = code2P->m_push;
			stack1P = code1P;
			stack2P = code2P;
		}
		for (best = weight; ++code1P < end_pcode1P && ++code2P < end_pcode2P;) {
			for (distance = 0; weight >= 0; weight += g_mismatch_cost) {
				at_code1P = code1P;
				at_code2P = code2P + distance;
				if (at_code2P >= end_pcode2P) {
					at_code1P += (at_code2P - end_pcode2P + 1);
					if (at_code1P >= end_pcode1P) {
						break;
					}
					at_code2P  = end_pcode2P - 1;
				} 
				at_code2P->m_matchP = 0;
				while (at_code2P >= code2P && at_code1P < end_pcode1P) {
					if (g_match_lines == 1) {
						if (stack1P < at_code1P) {
							++stack1P;
							assert(stack1P == at_code1P);
							stack1 -= stack1P->m_pop;
							if (stack1 < 0) {
								end_pcode1P = stack1P;
								break;
						}	}
						if (stack2P < at_code2P) {
							++stack2P;
							assert(stack2P == at_code2P);
							stack2 -= stack2P->m_pop;
							if (stack2 < 0) {
								end_pcode2P = stack2P;
							}
							goto move_back;
					}	}
					ret = same_pcode(at_code1P, at_code2P);
					if (ret > 0) {
						goto found;
					}
move_back:			++at_code1P;
					--at_code2P;
				}
				at_code1P[-1].m_matchP = 0;
				++distance;
			}
			break;
found:		weight += ret;
			if (weight >= best) {
				best       = weight;;
				end_code1P = at_code1P;
				end_code2P = at_code2P;
			}
			at_code1P->m_matchP = at_code2P;
			at_code2P->m_matchP = at_code1P;
			code1P              = at_code1P;
			code2P              = at_code2P;
		}
		
		++end_code1P;
		++end_code2P;
		if ( (((end_code1P - start_pcode1P) >= g_min_match_sequence) && 
			  ((end_code2P - start_pcode2P) >= g_min_match_sequence)
			 )
/*           || 
			(start_pcode1P == method1P->m_pcodeP && (end_code1P - start_pcode1P) == method1P->m_pcode_count) ||
			(start_pcode2P == method2P->m_pcodeP && (end_code2P - start_pcode2P) == method2P->m_pcode_count)
 */
		   ) {
			report_clone(end_code1P, end_code2P);
			for (at_code1P = start_pcode1P; at_code1P < end_code1P; ++at_code1P) {
				if ((at_code2P = at_code1P->m_matchP)) {
					if (matches_cacheP == matches_endP) {
						matches_cacheP     = (matchesT *) xmalloc(sizeof(matchesT) * 8192);
						matches_endP       = matches_cacheP + 8192;
					}
					matches_cacheP->m_nextP  = at_code1P->m_matchesP;
					matches_cacheP->m_pcodeP = at_code2P;
					at_code1P->m_matchesP    = matches_cacheP++;
			}	}
			return(end_code2P);
	}	}
	return(0);
}

void
detect(void)
{
	static	pcodeT			*start_pcodes2P, *end_pcodeP, *skipP;
	static	method_infoT	*method3P;
	static	matchesT		*matchesP;
	static	int				lineno;

	for (class1P = Cclass::g_headP; class1P; class1P = class1P->m_nextP) {
		if (g_verbose) {
			fprintf(stderr, "Matching %S\n", class1P->m_nameP);
		}
		method1P     = class1P->m_methodsP;
		for (end_method1P = method1P + class1P->m_methods_count; method1P < end_method1P; ++method1P) {
			start_pcode1P = method1P->m_pcodeP;
			for (end_pcodeP = start_pcode1P + method1P->m_pcode_count; start_pcode1P < end_pcodeP; ++start_pcode1P) {
				switch (g_match_lines) {
				case 2:
					lineno = start_pcode1P->m_lineno;
					if (lineno) {
						if (start_pcode1P->m_offset && lineno == start_pcode1P[-1].m_lineno) {
							continue;
						}
						break;
					}
				case 1:
					if (start_pcode1P->m_pop) {
						continue;
				}	}
					 
				for (start_pcode2P = start_pcode1P->m_hashP; start_pcode2P; ) {
					switch (g_match_lines) {
					case 2:
						lineno = start_pcode1P->m_lineno;
						if (lineno) {
							if (start_pcode2P->m_offset && lineno == start_pcode2P[-1].m_lineno) {
								goto advance;
							}
							break;
						}
					case 1:
						if (start_pcode2P->m_pop) {
							goto advance;
					}	}

					// Detect subsequence within larger sequences already matched
					for (matchesP = start_pcode1P->m_matchesP; matchesP; matchesP = matchesP->m_nextP) {
						if (matchesP->m_pcodeP == start_pcode2P) {
							goto advance;
					}	}

					start_pcodes2P = (start_pcode2P - start_pcode2P->m_offset);
					method2P       = ((method_infoT **) start_pcodes2P)[-1];
					end_pcode1P    = end_pcodeP;
					if (method2P == method1P) {
						if (start_pcode2P - start_pcode1P < g_min_match_sequence ||
							end_pcodeP    - start_pcode2P < g_min_match_sequence) {
							goto advance;
						}
						end_pcode1P = start_pcode2P;
						end_pcode2P = end_pcodeP;
					} else {
						end_pcode2P = start_pcodes2P + method2P->m_pcode_count;
					}
					class2P       = method2P->m_classP;
					++g_comparisons;
					skipP         = detect_clone();
					start_pcode2P = start_pcode2P->m_hashP;

					if (skipP) {
						for (; start_pcode2P; start_pcode2P = start_pcode2P->m_hashP) {
							method3P = ((method_infoT **) (start_pcode2P - start_pcode2P->m_offset))[-1];
							if (method3P != method2P || start_pcode2P >= skipP) {
								break;
					}	}	}
					continue;
advance:			start_pcode2P = start_pcode2P->m_hashP;
}	}	}	}	}
