#include "typedef.h"

#include <Windows.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "object.h"
#include "opcode.h"
#include "class.h"
#include "detect.h"

int		g_dynamic            = 0;
char	*g_html_dirP         = 0;
int		g_verbose            = 0;
int		g_report_java        = 1;
int		g_report_pcode       = 1;
int		g_match_lines		 = 1;
int		g_exceptions         = 1;
int		g_comments           = 0;
char	*g_classpathP        = 0;
char	*g_java[10]	         = {0,"-g","-classpath", ".", 0, 0};

int		g_min_match_sequence = 5;
int		g_mismatch_cost      = -1;
int		g_same_cost			 = 1;
int		g_comparisons		 = 0;

char	**g_envPP;

FILE	*stdoutF   = stdout;
FILE	*redirectF = stdoutF;

static void
print_title(int argc, char *argv[])
{
	int	i;

	fputs("<html><title>", stdoutF);
	fprintf(stdoutF, "JCD Version 1.0.4 Compiled " __DATE__ ":"  __TIME__ );
	fputs("</title>\n<body>\n", stdoutF);
	fprintf(stdoutF, "<h3>Arguments to %s</h3>\n<p>\n", argv[0]);
	for (i = 1; i < argc; ++i) {
		fprintf(stdoutF,"\"%s\" ", argv[i]);
	}
	fprintf(stdoutF, "<h3>Processing</h3>");
}

int
main(int argc, char *argv[], char *env[])
{
	static		int started = 0;
	char		*filenamesP = 0;
	FILE		*F;
	char		*P, *P1;
	int			i;
	time_t		start_clock, end_clock;
	
	start_clock = time(0);
	
	g_envPP = env;

	for (i = 1; i < argc;) {
		P = argv[i++];
		if (*P == '-') {
			switch (P[1]) {
			case 'a':
				g_match_lines = 0;
				if (i < argc) {
					P = argv[i++];
					switch (P[0]) {
					case 'l':
						g_match_lines = 1;	// Match start of java source lines
						break;
					case 's':
						g_match_lines = 2;	// Match start of pcode steps
						break;
				}	}
				break;
			case 'c':
				if (g_classpathP) {
					fprintf(stderr, "-c option may only be specified once\n");
					return(1);
				}
				if (i < argc) {
					g_classpathP = argv[i++];
					g_java[3]    = g_classpathP;
				}
				break;
			case 'd':
				g_dynamic = 1;
				break;
			case 'e':
				g_exceptions = 0;
				break;
			case 'i':
				if (filenamesP) {
					fprintf(stderr, "-i option may only be specified once\n");
					return(1);
				}
				if (i >= argc) {
					filenamesP = "-";
				} else {
					filenamesP = argv[i++];
				}
				break;
			case 'h':
				if (i < argc) {
					char	buffer[2048];

					g_html_dirP = argv[i++];
					CreateDirectoryA(g_html_dirP, 0);
					sprintf(buffer, "%s/index.html", g_html_dirP);
					redirectF = fopen(buffer, "w");
					if (!redirectF) {
						fprintf(stderr, "Can't open %s\n", buffer);
						return(1);
					}
					stdoutF = redirectF;
				}
				break;
			case 'j':
				if (i < argc) {
					if (g_java[0]) {
						fprintf(stderr, "-j/ava option may only be specified once\n");
						return(1);
					}
					g_java[0] = P1 = argv[i++];
				}
				break;
			case 'l':
				if (i < argc) {
					int val = atoi(argv[i++]);
					if (val < 1) {
						fprintf(stderr, "-l/ength must be > 0\n");
						return(1);
					}
					g_min_match_sequence = val;
				}
				break;
			case 'm':
				if (i < argc) {
					int val = atoi(argv[i++]);
					if (val < 1) {
						fprintf(stderr, "-m/ismatch cost must be > 0\n");
						return(1);
					}
					g_mismatch_cost = -1;
				}
				break;
			case 's':
				if (i < argc) {
					int val = atoi(argv[i++]);
					if (val < 1) {
						fprintf(stderr, "-s/ame cost must be > 0\n");
						return(1);
					}
					g_same_cost = -1;
				}
				break;
			case 'v':
				g_verbose = 1;
				break;
			case 'r':
				g_report_java = 0;
				break;
			case 'R':
				g_report_pcode = 0;
				break;
			case 'x':
				g_comments = 1;
				break;
			default:
				fputs("\n\nOptions:\n\n", stderr);
				fputs("  -a/ll [i/nstructions|l/ines|s/teps]\n", stderr);
				fputs("  -e/xceptions\n", stderr);
				fputs("  -j/ava <program>\n", stderr);
				fputs("  -c/lasspath <paths>\n", stderr);
				fputs("  -i/nput filenames\n", stderr);
				fputs("  -d/ynamic\n", stderr);
				fputs("  -h/tml <directory>\n", stderr);
				fputs("  -l/ength <min>\n", stderr);
				fputs("  -m/ismatch <cost>\n", stderr);
				fputs("  -s/ame <cost>\n", stderr);
				fputs("  -v/erbose\n", stderr);
				fputs("  -r/emove java reporting\n", stderr);
				fputs("  -R/emove pcode reporting\n", stderr);
				fputs("  -x/tra comments\n", stderr);
				return(1);
			}
			continue;
		}
		if (!started) {
			print_title(argc, argv);
			started = 1;
		}
		readclassfile(P, 0);
	}

	if (!started) {
		print_title(argc, argv);
		started = 1;
	}
	fputc('<p>\n', stdoutF);
	
	if (filenamesP) {
		char	filename[2048];
		char	*P;
		int		c;

		if (!strcmp(filenamesP, "-")) {
			F = stdin;
		} else {
			F = fopen(filenamesP, "r");
			if (!F) {
				fprintf(stderr, "Couldn't open %s\n", filenamesP);
				return(1);
		}	}
		do {
			P = filename;
			for (;;) {
				c = getc(F);
				if (c == '\r' || c == '\n' || c == EOF) {
					break;
				}
				*P++ = c;
			}
			if (P != filename) {
				*P = 0;
				readclassfile(filename, 0);
			}
		} while (c != EOF);

		if (F != stdin) {
			fclose(F);
	}	}
	// Now read recursively any classes which are cited but not yet processed
	if (g_dynamic) {
		fprintf(stdoutF, "<h3>Dynamic</h3>\n<p>\n");
		pseudo_source();
	}
	// Now perform clone detection

	start_report();
	detect();

	finalize_report();

	end_clock = time(0);
	fprintf(redirectF, "</table>\n<p>\n<table>\n<tr><td>Start</td><td>%s</td></tr>\n", 	ctime(&start_clock));
	fprintf(redirectF, "<tr><td>End</td><td>%s</td></tr>\n", 	ctime(&end_clock));
	fprintf(redirectF, "<tr><td>Comparisons</td><td align=right>%d</td></tr>\n", g_comparisons);
	fprintf(redirectF, "</table>\n<p>\n<i>End of Report</i>\n</body>\n</html>\n");

	return(0);
}

